#!/usr/bin/env python3
from __future__ import annotations
import logging
import numpy as np
import numpy.typing as npt
import pydantic
from copy import deepcopy
from ramanchada2.io.HSDS import write_cha, write_nexus
from ramanchada2.io.output.write_csv import write_csv as io_write_csv
from ramanchada2.misc.plottable import Plottable
from ramanchada2.misc.types import PositiveOddInt, SpeMetadataModel
from ramanchada2.misc.types.spectrum import SpeProcessingListModel, SpeProcessingModel
from scipy.signal import convolve, savgol_coeffs, savgol_filter
from scipy.stats import median_abs_deviation, rv_histogram
from typing import Dict, List, Set, Tuple, Union

logger = logging.getLogger(__name__)


class Spectrum(Plottable):
    _available_processings: Set[str] = set()

    @pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
    def __init__(self,
                 x: Union[npt.NDArray, int, None] = None,
                 y: Union[npt.NDArray, None] = None,
                 cachefile: str = '',
                 metadata: Union[SpeMetadataModel, None] = None,
                 applied_processings: Union[SpeProcessingListModel, None] = None):
        super(Plottable, self).__init__()
        if x is not None:
            if isinstance(x, int):
                self.x = np.arange(x)
            else:
                self.x = x
        if y is not None:
            self.y = y

        sort_idx = np.argsort(self.x)
        if (np.diff(sort_idx) != 1).any():
            self.x = self.x[sort_idx]
            if self.y is not None:
                self.y = self.y[sort_idx]

        self._cachefile = cachefile
        self._metadata = deepcopy(metadata or SpeMetadataModel(__root__={}))
        self._applied_processings = deepcopy(applied_processings or SpeProcessingListModel(__root__=[]))

    def __copy__(self):
        return Spectrum(
            x=self.x,
            y=self.y,
            cachefile=self._cachefile,
            metadata=self._metadata,
            applied_processings=self._applied_processings,
        )

    def __repr__(self):
        return self._applied_processings.repr()

    def __str__(self):
        return str(self._applied_processings.to_list())

    def write_csv(self, filename, delimiter=',', newline='\n'):
        csv = io_write_csv(self.x, self.y, delimiter=delimiter)
        with open(filename, 'w', newline=newline) as f:
            for c in csv:
                f.write(c)

    def write_cha(self, chafile, dataset):
        write_cha(chafile, dataset, self.x, self.y, self.meta.serialize())

    def write_nexus(self, chafile, dataset):
        write_nexus(chafile, dataset, self.x, self.y, self.meta.serialize())

    def write_cache(self):
        if self._cachefile:
            self.write_cha(
                self._cachefile,
                '/cache/'+self._applied_processings.cache_path()+'/_data')

    def process(self, algorithm: str, **kwargs):
        if algorithm not in self._available_processings:
            raise ValueError('Unknown algorithm {algorithm}')
        return getattr(self, algorithm)(**kwargs)

    @classmethod
    @pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
    def apply_creator(cls, step: SpeProcessingModel, cachefile_=None):
        proc = getattr(cls, step.proc)
        spe = proc(*step.args, **step.kwargs, cachefile_=cachefile_)
        return spe

    @pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
    def apply_processing(self, step: SpeProcessingModel):
        proc = getattr(self, step.proc)
        spe = proc(*step.args, **step.kwargs)
        return spe

    def _plot(self, ax, *args, **kwargs):
        ax.errorbar(
            self.x,
            self.y,
            xerr=self.x_err,
            yerr=self.y_err,
            **kwargs
        )

    def _sort_x(self):
        idx = np.argsort(self.x)
        self.x = self.x[idx]
        self.y = self.y[idx]

    @property
    def x(self): return np.array(self._xdata)

    @x.setter
    def x(self, val: npt.NDArray[np.float64]):
        self._xdata = val
        self._xdata.flags.writeable = False

    @property
    def x_bin_boundaries(self):
        return np.concatenate((
            [(3*self.x[0] - self.x[1])/2],
            (self.x[1:] + self.x[:-1])/2,
            [(3*self.x[-1] - self.x[-2])/2]
        ))

    @property
    def y(self) -> npt.NDArray[np.float64]:
        return np.array(self._ydata)

    @y.setter
    def y(self, val: npt.NDArray[np.float64]):
        self._ydata = val
        self._ydata.flags.writeable = False

    @property
    def y_noise(self):
        return self.y_noise_savgol()

    def y_noise_MAD(self):
        return median_abs_deviation(np.diff(self.y))

    def y_noise_savgol_DL(self, order: PositiveOddInt = PositiveOddInt(1)):
        npts = order + 2
        ydata = self.y - np.min(self.y)
        summ = np.sum((ydata - savgol_filter(ydata, npts, order))**2)
        coeff = savgol_coeffs(npts, order)
        coeff[(len(coeff) - 1) // 2] -= 1
        scale = np.sqrt(np.sum(coeff**2))
        return np.sqrt(summ/len(ydata))/scale

    @pydantic.validate_arguments
    def y_noise_savgol(self, order: PositiveOddInt = PositiveOddInt(1)):
        npts = order + 2

        # subtract smoothed signal from original
        coeff = - savgol_coeffs(npts, order)
        coeff[(len(coeff)-1)//2] += 1

        # normalize coefficients so that `sum(coeff**2) == 1`
        coeff /= np.sqrt(np.sum(coeff**2))

        # remove the common floor
        ydata = self.y - np.min(self.y)
        return np.std(convolve(ydata, coeff, mode='same'))

    @property
    def x_err(self):
        return np.zeros_like(self._xdata)

    @property
    def y_err(self):
        return np.zeros_like(self._ydata)

    @property
    def meta(self) -> SpeMetadataModel:
        return self._metadata

    @meta.setter
    @pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
    def meta(self, val: Union[Dict, SpeMetadataModel]):
        if isinstance(val, dict):
            self._metadata = SpeMetadataModel.parse_obj(val)
        else:
            self._metadata = val

    @property
    def result(self):
        return self.meta['ramanchada2_filter_result']

    @result.setter
    def result(self, res: Union[Dict, List]):
        return self.meta._update(dict(ramanchada2_filter_result=res))

    @pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
    def spe_distribution(self, trim_range: Union[Tuple[float, float], None] = None):
        x_all = self.x_bin_boundaries
        if trim_range is not None:
            l_idx = int(np.argmin(np.abs(x_all - trim_range[0])))
            r_idx = int(np.argmin(np.abs(x_all - trim_range[1])))
            spe_dist = rv_histogram((self.y[l_idx:r_idx], x_all[l_idx:r_idx+1]))
        else:
            spe_dist = rv_histogram((self.y, x_all))
        return spe_dist

    @pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
    def gen_samples(self, size: pydantic.PositiveInt, trim_range=None):
        spe_dist = self.spe_distribution(trim_range=trim_range)
        samps = spe_dist.rvs(size=size)
        return samps



#!/usr/bin/env python3

from typing import Dict, List, Literal, Union

import numpy as np
import numpy.typing as npt
from lmfit import lineshapes
from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_constructor

from ..spectrum import Spectrum


@add_spectrum_constructor()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def from_theoretical_lines(
        shapes: List[Literal[lineshapes.functions]],  # type: ignore
        params: List[Dict],
        x: Union[int, npt.NDArray[np.float64]] = 2000):
    """
    Generate spectrum from `lmfit` shapes.

    Args:
        shapes:
            The shapes to be used for spectrum generation.
        params:
            Shape parameters to be applied to be used with shapes.
        x:
            Array with `x` values, by default `np.array(2000)`.
    """
    spe = Spectrum(x=x)
    x = spe.x
    y = np.zeros_like(x, dtype=float)
    for shape_name, pars in zip(shapes, params):
        shape = getattr(lineshapes, shape_name)
        y += shape(x=x, **pars)
    spe.y = y
    return spe



#!/usr/bin/env python3

from typing import TextIO, Tuple, Dict

from numpy.typing import NDArray

from .bw_format import bw_format
from .two_column_spe import two_column_spe


""" There are 4 types of TXT data files that can be distinguished by their first line:
    1. File Version;BWRam4.11_11
    2. File Version;BWSpec4.11_1
    3. <wavenumber>	<intensity>
    4. #Wave		#Intensity
      <wavenumber>	<intensity>
    """


def read_txt(data_in: TextIO) -> Tuple[NDArray, NDArray, Dict]:
    lines = data_in.readlines()
    if lines[0].startswith('File Version;BW'):
        data, meta = bw_format(lines)
        positions = data['RamanShift'].to_numpy()
        intensities = data['DarkSubtracted#1'].to_numpy()
    else:  # assume two column spectrum
        positions, intensities = two_column_spe(lines)
        meta = dict()
    return positions, intensities, meta



#!/usr/bin/env python3


from .metadata import SpeMetadataModel
from .metadata import SpeMetadataTyping
from .metadata import SpeMetadataFieldTyping
from .applied_processings import SpeProcessingModel
from .applied_processings import SpeProcessingListModel



#!/usr/bin/env python3

from .spectrum import Spectrum  # noqa
from .arithmetics import *  # noqa
from .baseline import *  # noqa
from .calc import *  # noqa
from .calibration import *  # noqa
from .filters import *  # noqa
from .peaks import *  # noqa
from .creators.from_cache_or_calc import from_cache_or_calc
from .creators.from_chada import from_chada
from .creators.from_local_file import from_local_file
from .creators.from_simulation import from_simulation
from .creators.from_theoretical_lines import from_theoretical_lines
from .creators.from_spectral_component_collection import from_spectral_component_collection
from .creators.from_delta_lines import from_delta_lines
from .creators.from_test_spe import from_test_spe
from .creators.hdr_from_multi_exposure import hdr_from_multi_exposure



#!/usr/bin/env python3

from typing import Literal, Union, Callable

import numpy as np
import numpy.typing as npt
from numpy.typing import NDArray
from scipy import signal
import lmfit
from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def convolve(
        old_spe: Spectrum,
        new_spe: Spectrum, /,
        lineshape: Union[Callable[[Union[float, NDArray]], float],
                         npt.NDArray,
                         Literal[
                              'gaussian', 'lorentzian',
                              'voigt', 'pvoigt', 'moffat',
                              'pearson4', 'pearson7'
                              ]],
        **kwargs):
    """
    Convole spectrum with arbitrary lineshape.

    Args:
        lineshape:
            Can be a callable, `str` or `np.ndarray`. If callable: should have a single positional argument `x`, e.g.
            `lambda x: np.exp((x/5)**2)`. If predefined peak profile: can be `gaussian`, `lorentzian`, `voigt`,
            `pvoigt`, `moffat` or `pearson4`. If `np.ndarray`: lineshape in samples.
        **kwargs:
            Additional kwargs will be passed to lineshape function.
    """

    if isinstance(lineshape, np.ndarray):
        new_spe.y = signal.convolve(old_spe.y, lineshape, mode='same')
    else:
        if callable(lineshape):
            shape_fun = lineshape
        else:
            shape_fun = getattr(lmfit.lineshapes, lineshape)

        leny = len(old_spe.y)
        x = np.arange(-(leny-1)//2, (leny+1)//2, dtype=float)
        shape_val = shape_fun(x, **kwargs)
        new_spe.y = signal.convolve(old_spe.y, shape_val, mode='same')



#!/usr/bin/env python3

from typing import List
import numpy as np
from pydantic import validate_arguments, confloat, PositiveInt
from scipy import signal, fft
import emd

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def derivative_sharpening(old_spe: Spectrum,
                          new_spe: Spectrum, /,
                          filter_fraction: confloat(gt=0, le=1) = .6,  # type: ignore
                          sig_width: confloat(ge=0) = .25,  # type: ignore
                          der2_factor: float = 1,
                          der4_factor: float = .1
                          ):
    leny = len(old_spe.y)
    Y = fft.rfft(old_spe.y, n=leny)
    h = signal.windows.hann(int(len(Y)*filter_fraction))
    h = h[(len(h))//2-1:]
    h = np.concatenate((h, np.zeros(len(Y)-len(h))))
    der = np.arange(len(Y))
    der = 1j*np.pi*der/der[-1]
    Y *= h
    Y2 = Y*der**2
    Y4 = Y2*der**2
    y0 = fft.irfft(Y, n=leny)
    y2 = fft.irfft(Y2, n=leny)
    y4 = fft.irfft(Y4, n=leny)
    new_spe.y = y0 - y2/sig_width**2*der2_factor + y4/sig_width**4*der4_factor


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def hht_sharpening(old_spe: Spectrum,
                   new_spe: Spectrum, /,
                   movmin=100
                   ):
    imfs = emd.sift.sift(old_spe.y).T
    freq_list = list()
    for ansig in signal.hilbert(imfs):
        freq_list.append(emd.spectra.freq_from_phase(
            emd.spectra.phase_from_complex_signal(ansig, ret_phase='unwrapped'), 1))
    freq = np.array(freq_list)
    freq[freq < 0] = 0
    freq[np.isnan(freq)] = 0
    imfsall = imfs.copy()
    imfsall[np.isnan(imfsall)] = 0
    imfsall[freq > .3] = 0
    imfsall *= freq**.5
    ynew = np.sum(imfsall, axis=0)
    new_spe.y = ynew
    new_spe.y = new_spe.subtract_moving_minimum(movmin).normalize().y  # type: ignore
    new_spe.y = new_spe.y * old_spe.y.max() + old_spe.y.min()


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def hht_sharpening_chain(old_spe: Spectrum,
                         new_spe: Spectrum, /,
                         movmin: List[PositiveInt] = [150, 50]
                         ):
    spe = old_spe
    for mm in movmin:
        spe = spe.hht_sharpening(movmin=mm)  # type: ignore
    new_spe.y = spe.y



#!/usr/bin/env python3

from __future__ import annotations

from typing import Dict, List

import pydantic


class SpeProcessingModel(pydantic.BaseModel):
    proc: str = pydantic.Field(...)
    args: List = list()
    kwargs: Dict = dict()

    @property
    def is_constructor(self):
        from ramanchada2.misc.spectrum_deco.dynamically_added import dynamically_added_constructors
        return self.proc in dynamically_added_constructors

    @pydantic.validator('proc', pre=True)
    @pydantic.validate_arguments
    def check_proc(cls, val: str):
        from ramanchada2.misc.spectrum_deco.dynamically_added import (dynamically_added_filters,
                                                                      dynamically_added_constructors)
        if val in dynamically_added_filters:
            return val
        if val in dynamically_added_constructors:
            return val
        raise ValueError(f'processing {val} not supported')


class SpeProcessingListModel(pydantic.BaseModel):
    __root__: List[SpeProcessingModel]

    def __len__(self):
        return len(self.__root__)

    def append(self, proc, args=[], kwargs={}):
        self.__root__.append(SpeProcessingModel(proc=proc, args=args, kwargs=kwargs))

    @pydantic.validate_arguments
    def extend_left(self, proc_list: List[SpeProcessingModel]):
        self.__root__ = proc_list + self.__root__

    def pop(self):
        return self.__root__.pop()

    def clear(self):
        return self.__root__.clear()

    def assign(self, *args, **kwargs):
        self.clear()
        self.append(*args, **kwargs)

    def _string_list(self):
        ret = list()
        for elem in self.__root__:
            args = [f'{repr(i)}' for i in elem.args]
            kwargs = [f'{k}={repr(v)}' for k, v in elem.kwargs.items()]
            comb = ', '.join(args + kwargs)
            ret.append(f'{elem.proc}({comb})')
        return ret

    def repr(self):
        return '.'.join(self._string_list())

    def cache_path(self):
        return '/'.join([
            i.replace(' ', '').replace('/', '_')
            for i in self._string_list()])

    def to_list(self):
        return [i.dict() for i in self.__root__]



#!/usr/bin/env python3

import os
import glob

__all__ = [
    os.path.basename(f)[:-3]
    for f in glob.glob(os.path.dirname(__file__)+"/*.py")
    if os.path.isfile(f) and not os.path.basename(f).startswith('_')
]



import pandas as pd
import importlib.resources
from functools import reduce

data = [
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/NeonSNQ043_iR532_Probe_5msx2.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'Neon'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/NeonSNQ043_iR532_Probe_100msx2.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'Neon'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/PST10_iR532_Probe_100_3000msx7.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'PST'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/S0B10_iR532_Probe_100_60000msx2.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'S0B'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/S0N10_iR532_Probe_100_30000msx3.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'S0N'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/S0P10_iR532_Probe_100_60000msx2.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'S0P'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/S1N10_iR532_Probe_100_22000msx2.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'S1N'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/Sil10_iR532_Probe_100_60000msx2.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'Sil_'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/nCAL10_iR532_Probe_100_2500msx3.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'nCAL'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/sCAL10_iR532_Probe_100_3200msx4.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'sCAL'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/LED532_Probe_40msx3_1.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'LED532_EL0-9001'},
    {'OP': '01', 'device': 'BWtek', 'filename': './FMNT-M_BW532/NIST532_Probe_3000msx8_1.txt',
        'laser_wl': '532', 'provider': 'FNMT', 'sample': 'NIST532_SRM2242a'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/NeonSNQ043_iR785_OP01.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'Neon'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/PST10_iR785_OP01_40000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'PST'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0B10_iR785_OP01_6000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0B'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0N10_iR785_OP01_6000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0N'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0P10_iR785_OP01_6000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0P'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S1N10_iR785_OP01_6000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S1N'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/Sil10_iR785_OP01_6000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'Sil'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/nCAL10_iR785_OP01_6000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'nCAL'},
    {'OP': '01', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/sCAL10_iR785_OP01_4000msx4.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'sCAL'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/NeonSNQ043_iR785_OP02.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'Neon'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/PST10_iR785_OP02_50000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'PST'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0B10_iR785_OP02_25000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0B'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0N10_iR785_OP02_25000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0N'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0P10_iR785_OP02_25000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0P'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S1N10_iR785_OP02_25000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S1N'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/Sil10_iR785_OP02_25000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'Sil'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/nCAL10_iR785_OP02_20000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'nCAL'},
    {'OP': '02', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/sCAL10_iR785_OP02_15000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'sCAL'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/NeonSNQ043_iR785_OP03.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'Neon'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/PST10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'PST'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0B10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0B'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0N10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0N'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S0P10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S0P'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/S1N10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'S1N'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/Sil10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'Sil'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/nCAL10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'nCAL'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/sCAL10_iR785_OP03_8000msx2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'sCAL'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/LED785_Lens_1000x10_2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'NIR785_EL0-9002'},
    {'OP': '03', 'device': 'Horiba', 'filename': './FMNT-M_Ho785/NIST785_Lens_80000x5_2.txt',
        'laser_wl': '785', 'provider': 'FNMT', 'sample': 'NIST785_SRM2241'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/Ne_532nm_x50_800ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'Neon'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/Ne_532nm_x50_25ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'Neon'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/PST02_iRPlus532_Z050_100_2500msx5.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'PST'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/S0B02_iRPlus532_Z050_100_30000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S0B'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/S0N02_iRPlus532_Z050_100_40000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S0N'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/S0P02_iRPlus532_Z050_100_30000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S0P'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/S1N02_iRPlus532_Z050_100_12000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S1N'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/nCAL02_iRPlus532_Z050_100_9000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'nCAL'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW532/sCAL02_iRPlus532_Z050_100_20000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'sCAL'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/Ne_532nm_x100_2000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'Neon'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/Ne_532nm_x100_110ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'Neon'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/PST02_iRPlus532_Z100_100_50000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'PST'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/S0B02_iRPlus532_Z100_100_22000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S0B'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/S0N02_iRPlus532_Z100_100_28000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S0N'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/S0P02_iRPlus532_Z100_100_20000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S0P'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/S1N02_iRPlus532_Z100_100_6500ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'S1N'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/nCAL02_iRPlus532_Z100_100_30000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'nCAL'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/sCAL02_iRPlus532_Z100_100_65000ms.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'sCAL'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/Visible_532nm_130ms_100X_2.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'LED532_EL0-9001'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW532/NISTSRM2242aC_BW532_100x_800msx10.txt',
        'laser_wl': '532', 'provider': 'ICV', 'sample': 'NIST532_SRM2242a'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/PST02_iRPlus785_Z020_100_1300ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'PST'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/S0B02_iRPlus785_Z020_100_full.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0B'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/S0B02_iRPlus785_Z020_100_12000ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0B'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/S0N02_iRPlus785_Z020_100_3800ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0N'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/S0P02_iRPlus785_Z020_100_12000ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0P'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/S1N02_iRPlus785_Z020_100_3800ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S1N'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/nCAL02_iRPlus785_Z020_100_1100ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'nCAL'},
    {'OP': '020', 'device': 'BWtek', 'filename': './ICV_BW785/sCAL02_iRPlus785_Z020_100_2400ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'sCAL'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW785/PST02_iRPlus785_Z050_100_3200ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'PST'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW785/S0B02_iRPlus785_Z050_100_5500ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0B'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW785/S0N02_iRPlus785_Z050_100_3200ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0N'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW785/S0P02_iRPlus785_Z050_100_5700ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0P'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW785/S1N02_iRPlus785_Z050_100_2100ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S1N'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW785/nCAL02_iRPlus785_Z050_100_2100ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'nCAL'},
    {'OP': '050', 'device': 'BWtek', 'filename': './ICV_BW785/sCAL02_iRPlus785_Z050_100_3800ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'sCAL'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/PST02_iRPlus785_Z100_100_9000ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'PST'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/S0B02_iRPlus785_Z100_100_4700ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0B'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/S0N02_iRPlus785_Z100_100_3100ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0N'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/S0P02_iRPlus785_Z100_100_4800ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S0P'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/S1N02_iRPlus785_Z100_100_2000ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'S1N'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/nCAL02_iRPlus785_Z100_100_6000ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'nCAL'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/sCAL02_iRPlus785_Z100_100_9500ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'sCAL'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/NIR_785nm_2000ms_100X_2.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'NIR785_EL0-9002'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/NISTSRM2241_BW785_100x_25sx5.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'NIST785_SRM2241'},
    {'OP': '100', 'device': 'BWtek', 'filename': './ICV_BW785/Ne_785nm_x20_50ms.txt',
        'laser_wl': '785', 'provider': 'ICV', 'sample': 'Neon'},

    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/neon_new2_Z010.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'Neon'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/Pol_HLR633_Z010_100_15sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'PST'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/S0B_HLR633_Z010_100_40sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'S0B'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/S0N_HLR633_Z010_100_15sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'S0N'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/S0P_HLR633_Z010_100_40sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'S0P'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/S1N_HLR633_Z010_100_10sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'S1N'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/Si_HLR633_Z010_100_40sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'Sil'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/nCAL_HLR633_Z010_100_5sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'nCAL'},
    {'OP': '010', 'device': 'Horiba', 'filename': './TOP_Ho633/sCAL_HLR633_Z010_100_15sx5.txt',
        'laser_wl': '633', 'provider': 'TOP', 'sample': 'sCAL'},
]


df = pd.DataFrame.from_dict(data)
loc = importlib.resources.files(__name__)


def filtered_df(**kwargs):
    filters = [reduce(lambda a, b: a | b, [df[k] == i for i in v]) for k, v in kwargs.items() if v]
    if filters:
        return df[reduce(lambda a, b: a & b, filters)]
    else:
        return df


def get_filters(**kwargs):
    df_tmp = filtered_df(**kwargs)
    return {c: df_tmp[c].unique() for c in df_tmp.columns if c != 'filename'}


def get_filenames(**kwargs):
    df_tmp = filtered_df(**kwargs)
    return df_tmp['filename']


def prepend_prefix(filenames):
    return [str(loc.joinpath(fn)) for fn in filenames]



#!/usr/bin/env python3

from typing import Callable, Union

import numpy as np
import numpy.typing as npt
from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def scale_xaxis_linear(old_spe: Spectrum,
                       new_spe: Spectrum, /,
                       factor: float = 1,
                       preserve_integral: bool = False):
    new_spe.x = old_spe.x * factor
    if preserve_integral:
        new_spe.y = old_spe.y / factor


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def scale_xaxis_fun(old_spe: Spectrum,
                    new_spe: Spectrum, /,
                    fun: Callable[[Union[int, npt.NDArray]], float],
                    args=[]):
    new_spe.x = fun(old_spe.x, *args)
    if (np.diff(new_spe.x) < 0).any():
        raise ValueError('The provided function is not a monoton increasing funciton.')



#!/usr/bin/env python3

from typing import List, Tuple, Dict

import pandas


def bw_format(lines: List[str]) -> Tuple[pandas.DataFrame, Dict]:
    section_split = 0
    for i, ll in enumerate(lines):
        if ll.count(';') > 1:
            section_split = i
            break
    meta = lines[:section_split]
    spec = lines[section_split:]

    meta_dict = dict([m.replace(',', '.').strip().split(';') for m in meta])

    spec_split = [s.replace(',', '.').replace(' ', '').strip('\r\n ;').split(';') for s in spec]
    spe_parsed = pandas.DataFrame.from_records(data=spec_split[1:], columns=spec_split[0])
    spe_parsed = spe_parsed.apply(pandas.to_numeric)
    spe_parsed = spe_parsed.dropna(axis=0)

    return spe_parsed, meta_dict



import numpy as np
from pydantic import PositiveInt, validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_filter

from ..spectrum import Spectrum


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _moving_median(s,
                   window_size: PositiveInt = 10):
    y = ([np.median(s[:window_size]) for i in range(window_size)] +
         [np.median(s[i-window_size: i+window_size]) for i in range(window_size, len(s)-window_size)] +
         [np.median(s[-window_size:]) for i in range(window_size)]
         )
    return np.array(y)


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def moving_median(old_spe: Spectrum,
                  new_spe: Spectrum, /,
                  window_size: PositiveInt = 10):
    """
    Moving median filter.

    Args:
        window_size:
            `int`, optional, default is `10`.
    """

    new_spe.y = _moving_median(old_spe.y, window_size)


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def subtract_moving_median(
        old_spe: Spectrum,
        new_spe: Spectrum,
        window_size: int):
    new_spe.y = old_spe.y - _moving_median(old_spe.y, window_size)



#!/usr/bin/env python3

from typing import List, Tuple

import numpy as np
from numpy.typing import NDArray


def two_column_spe(lines: List[str]) -> Tuple[NDArray, NDArray]:
    """
    Parse two column spectrum.

    Args:
        lines:
            List of lines to parse.

    Returns:
        Tuple[NDArray, NDArray]: positions and intensities.
    """
    positions, intensities = np.genfromtxt(lines).T
    return positions, intensities



import ast
from os import path

from setuptools import find_packages, setup

NAME = 'ramanchada2'

with open(path.join(path.dirname(__file__), 'src/ramanchada2/__init__.py')) as fd:
    VERSION = [expr.value.value
               for expr in ast.parse(fd.read()).body
               if (isinstance(expr, ast.Assign) and
                   len(expr.targets) == 1 and
                   expr.targets[0].id == '__version__')
               ][-1]

DESCRIPTION = 'Harmonising Raman Spectroscopy'
README_FILE = path.join(path.dirname(__file__), 'README.pypi')
LONG_DESCRIPTION = open(README_FILE, 'r', encoding='utf-8').read()
LONG_DESCRIPTION_CONTENT_TYPE = 'text/markdown'
URL = 'https://github.com/h2020charisma/ramanchada2'
AUTHOR = 'IDEAconsult Ltd.'
AUTHOR_EMAIL = 'dev-charisma@ideaconsult.net'
LICENSE = 'MIT'

KEYWORDS = [
    'Raman',
    'spectroscopy',
]

PYTHON_REQUIRES = '>=3.9'

PACKAGES = find_packages(where='src')

PACKAGE_DIR = {'': 'src'}

PACKAGE_DATA = {'': ['auxiliary/**/*.txt']}

DATA_FILES = []

INSTALL_REQUIRES = [
        "brukeropusreader",  # rc1-parser
        "h5py",
        "lmfit",
        "matplotlib",
        "numpy",
        "pandas",
        "pydantic==1.*",
        "emd",
        "renishawWiRE",  # rc1-parser
        "scikit-learn",
        "scipy>=1.8.0",
        "spc-io~=0.0.2",  # rc1-parser
        "statsmodels",
        "uncertainties",
]

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Libraries',
]


def setup_package():
    setup(
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        classifiers=CLASSIFIERS,
        data_files=DATA_FILES,
        description=DESCRIPTION,
        install_requires=INSTALL_REQUIRES,
        keywords=KEYWORDS,
        license=LICENSE,
        long_description=LONG_DESCRIPTION,
        long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
        name=NAME,
        package_dir=PACKAGE_DIR,
        package_data=PACKAGE_DATA,
        include_package_data=True,
        packages=PACKAGES,
        python_requires=PYTHON_REQUIRES,
        url=URL,
        version=VERSION,
    )


if __name__ == '__main__':
    setup_package()



import numpy as np

from renishawWiRE import WDFReader
from spc_io import SPC
from brukeropusreader import read_file


def readWDF(file):
    s = WDFReader(file)
    y_data = s.spectra
    x_data = s.xdata
    if np.mean(np.diff(x_data)) < 0:
        y_data = np.flip(y_data)
        x_data = np.flip(x_data)
    static_metadata = {
        "laser wavelength": s.laser_length,
        "no. of accumulations": s.accumulation_count,
        "spectral unit": s.spectral_unit.name,
        "OEM software name": s.application_name,
        "OEM software version": s.application_version
        }
    return x_data, y_data, static_metadata


def readSPC(file):
    with open(file, 'rb') as f:
        spc = SPC.from_bytes_io(f)
    if len(spc) != 1:
        raise ValueError(f'Only 1 sub supported, {len(spc)} found')
    x_data = spc[0].xarray
    y_data = spc[0].yarray
    static_metadata = spc.log_book.text
    return x_data, y_data, static_metadata


def readOPUS(file, obj_no=0):
    opus_data = read_file(file)
    x = opus_data.get_range("AB")
    y = opus_data["AB"]
    meta = {}
    for key in opus_data:
        if key == "AB":
            continue
        if isinstance(opus_data[key], dict):
            for subkey in opus_data[key]:
                meta["{}.{}".format(key, subkey)] = opus_data[key][subkey]
        else:
            meta[key] = opus_data[key]
    return x, y, meta



#!/usr/bin/env python3

from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def scale_yaxis_linear(old_spe: Spectrum,
                       new_spe: Spectrum,
                       factor: float = 1):
    new_spe.y = old_spe.y * factor



from .baseline_base import BaseLineBase


class BaselineNumerical(BaseLineBase):
    def __init__(self, x, y):
        """
        Numerical baseline

        Args:
            x:
                array-like
            y:
                array-like
        """



import os.path
from .third_party_readers import readSPC, readWDF, readOPUS
from .txt_format_readers import read_JCAMP, readTXT
from .binary_readers import readSPA, read_ngs


class UnsupportedFileTypeError(Exception):
    pass


def parse(source_path, file_type=None):
    filename, file_extension = os.path.splitext(source_path)

    if file_type is None:
        file_type = file_extension[1:].lower()  # without the leading dot

    if file_type in {'spc', 'sp'}:
        reader = readSPC
    elif file_type in {'spa'}:
        reader = readSPA
    elif file_type in {'0', '1', '2'}:
        reader = readOPUS
    elif file_type in {'wdf'}:
        reader = readWDF
    elif file_type in {'ngs'}:
        reader = read_ngs
    elif file_type in {'jdx', 'dx'}:
        reader = read_JCAMP
    elif file_type in {'txt', 'txtr', 'csv', 'prn', 'rruf'}:
        reader = readTXT
    else:
        raise UnsupportedFileTypeError(
            f'file type "{file_type}" is not supported'
        )

    x_data, y_data, metadata = reader(source_path)
    # Get rid of bytes that are found in some of the formats
    metadata = cleanMeta(metadata)
    # Flatten metadata
    metadata = dict(zip(metadata.keys(), [str(v) for v in metadata.values()]))
    # Extract metadata from native metadata and spectrum data,
    # store in metadata dictionary, and include in CHADA archive.
    metadata["Original file"] = os.path.basename(source_path)
    return x_data, y_data, metadata


def cleanMeta(meta):
    # This cleans complex-strcutures metadata, and returns a dict
    if isinstance(meta, dict):
        meta = {i: meta[i] for i in meta if i != ""}
        for key, value in meta.items():
            meta[key] = cleanMeta(value)
    if isinstance(meta, list):
        for ii, value in enumerate(meta):
            meta[ii] = cleanMeta(value)
    if isinstance(meta, str):
        meta = meta.replace('\\x00', '')
        meta = meta.replace('\x00', '')
    if isinstance(meta, bytes):
        try:
            meta = meta.decode('utf-8')
            meta = cleanMeta(meta)
        except Exception:
            meta = []
    return meta
